{$R+,S+,I+,Q+}
{ $DEFINE debug}
{
  ***********************************************************
  * Gerd's AVR-Assembler Version 1.3 as of April 2005       *
  * Free Pascal, source code, (C)2005 by DG4FAC             *
  * New in Version 1.3: See ReadMe.Txt for version info     *
  ***********************************************************
}
Program gavrasm;

Uses Crt,SysUtils,gavrinst,gavriset,gavrout,gavrtype,gavrmacr,gavrsymb,
  gavrline,gavrlang,gavrdefs,gavrif;

Const
  nHeader=3;
  asHeader:Array[1..nHeader] Of String=(
    '+------------------------------------------------------------+',
    '| gavrasm gerd''s AVR assembler Version 1.3 (C)2005 by DG4FAC |',
    '+------------------------------------------------------------+');
  lHeader:String='gavrasm Gerd''s AVR assembler version 1.3 (C)2005 by DG4FAC';
  cTab=Char(9);
  iMaxInt:LongInt=$3FFFFFFF; { 1073741823 }

Var sfn:String; { filename of the source file }
    sListFile,sEepFile,sCodeFile,sXErrorFile:String; { Result files }
    fListFileOpen:Boolean; { List File is open }
    fQuiet,fSymbols,fList,fListOff,fAnsiOff,fErrLong,fWrapOn,fWrap,
    fTypesList,fOptionList,fListMac,fBeginner,fListDirectives:Boolean; { Command line options }
    fl,fx:Text; { list-file, error-file }
    spFile:String; { Name of currently processed file }
    fExit:Boolean; { Skips further processing of the source file }
    fExitFile:Boolean; { Skips further processing of the file }
    fMacroDef:Boolean; { Inside macro definition }
    npLine:LongInt; { Currently processed line number }
    nerr,nw:LongInt; { Number of errors and s per pass }
    pc,pe,pd:LongInt; { Current PC/EEPROM/SRAM count }
    ncode,nconstants,ne,nd:LongInt; { Number of code/EEPROM/SRAM created }
    cSegm:Char; { Segment target, C/E/D }
    pass:Byte; { Current pass number }
    k:Byte; { For header output }
    w1,w2:Word; { Command words }
    sListLine2:String; { Second list line }

Function ProcessFile(sf:String):Boolean; Forward;

{ ------------------------- }
{ Read command line options }
{ ------------------------- }
Function GetOptions:Boolean;
Var j,k:LongInt;
  s:String;
Begin
GetOptions:=True;
fList:=False;
fQuiet:=False;
fSymbols:=False;
fTypesList:=False;
fListMac:=False;
fListDirectives:=False;
fErrLong:=False;
fBeginner:=False;
sfn:='test.asm';
For j:=1 To ParamCount Do
  Begin
  s:=ParamStr(j);
  If s[1]='-' Then
    Begin
    For k:=2 To Length(s) Do
      Begin
      Case UpCase(s[k]) Of
        '?','H':fOptionList:=True;
        'A':fAnsiOff:=Not fAnsiOff;
        'B':fBeginner:=True;
        'D':fListDirectives:=True;
        'E':fErrLong:=True;
        'L':fList:=True;
        'M':fListMac:=True;
        'Q':fQuiet:=True;
        'S':fSymbols:=True;
        'T':fTypesList:=True;
        'W':fWrap:=True;
        Else
        Writeln(GetMsgM(1),s[k],'!');
        GetOptions:=False;
        End;
      End;
    End Else
    Begin
    sfn:=s;
    End;
  End;
If fTypesList Or fOptionList Or fListDirectives Then
  Begin
  GetOptions:=False;
  If fTypesList Then ListAvrTypes;
  If fListDirectives Then ListDirectives;
  If fOptionList Then
    Begin
    Writeln;
    Writeln('Calls: gavrasm [-ABELQSW] source[.asm]');
    Writeln('       gavrasm [-?DHT]');
    Writeln;
    Writeln('List of options:');
    Writeln('  -L Listfile off            -M List macros');
    Writeln('  -Q Quiet, no screen output -S Symbol list in listfile');
    Writeln('  -B Beginners error comment -A Toggle ANSI output on command line');
    Writeln('  -E Longer error comments   -D List all supported directives');
    Writeln('  -W Enable wrapping         -T List all supported AVR types');
    Writeln('  -H -? List options (this)');
    Writeln;
    End;
  End Else
  Begin
  If ExtractFileExt(sfn)='' Then sfn:=sfn+'.asm';
  If FileExists(sfn) Then
    Begin
    sListFile:=ChangeFileExt(sfn,'.lst');
    sEepFile:=ChangeFileExt(sfn,'.eep');
    sCodeFile:=ChangeFileExt(sfn,'.hex');
    sXErrorFile:=ChangeFileExt(sfn,'.err');
    End Else
    Begin
    GetOptions:=False;
    Writeln(GetMsgM(2),sfn,'!');
    End;
  End;
fListOff:=fList;
End;

{ ----------------------------- }
{ Get date and time for listing }
{ ----------------------------- }

Function GetDateAndTime:String;
Begin
GetDateAndTime:=FormatDateTime('dd.mm.yyyy", "hh:nn:ss',Now);
End;

{ --------------------------- }
{ Process errors and warnings }
{ --------------------------- }

Procedure Dbg(s:String); { Internal compiler errors }
Begin
Inc(nerr);
Writeln;
Writeln(GetMsgM(59));
Writeln(GetMsgM(6),cl.sLine);
Writeln(' ==> ',s);
Writeln;
If Not fListOff Then
  Begin
  Writeln(fl);
  Writeln(fl,GetMsgM(59));
  Writeln(fl,GetMsgM(6),cl.sLine);
  Writeln(fl,' ==> ',s);
  End;
End;

Procedure BeginnersError; { list syntax for the instructions }
Begin
If Not fListOff Then
  Begin
  Write(fl,'      Syntax: ');
  If cl.nDirective<>0 Then
    Begin
    If cl.nDirective>0 Then Writeln(fl,GetMsgM(61+cl.nDirective)) Else
      Writeln(fl,GetMsgM(60));
    End Else
    If Inst.sM<>'' Then
    Begin
    If Inst.c1=' ' Then
      Writeln(fl,Inst.sM,GetMsgM(40)) Else
      Begin
      If Inst.c2=' ' Then
        Begin
        Writeln(fl,Inst.sM,' ',GetMsgM(58));
        Write(fl,'      ',GetMsgM(58),': ');
        Case Inst.c1 Of
          'R':Writeln(fl,GetMsgM(41));
          'B':Writeln(fl,GetMsgM(42));
          '1':Writeln(fl,GetMsgM(43));
          'X':Writeln(fl,GetMsgM(44));
          'E':Writeln(fl,GetMsgM(45));
          '2':Writeln(fl,GetMsgM(46));
          'H':Writeln(fl,GetMsgM(47));
          Else
          Dbg('Illegal c1 ('+Inst.c1+') in single parameter instruction!');
          End;
        End Else
        Begin
        Writeln(fl,Inst.sM,' ',GetMsgM(58),'1,',GetMsgM(58),'2');
        Write(fl,'      ',GetMsgM(58),' 1: ');
        Case Inst.c1 Of
          'R':Writeln(fl,GetMsgM(41));
          'D':Writeln(fl,GetMsgM(48));
          'H':Writeln(fl,GetMsgM(47));
          'B':Writeln(fl,GetMsgM(42));
          'Q':Writeln(fl,GetMsgM(49));
          'J':Writeln(fl,GetMsgM(50));
          'K':Writeln(fl,GetMsgM(51));
          'P':Writeln(fl,GetMsgM(52));
          'S':Writeln(fl,GetMsgM(53));
          'T':Writeln(fl,GetMsgM(54));
          'W':Writeln(fl,GetMsgM(55));
          Else
          Dbg('Illegal c1 ('+Inst.c1+') in double parameter instruction!');
          End;
        Write(fl,'      ',GetMsgM(58),' 2: ');
        Case Inst.c2 Of
          'R':Writeln(fl,GetMsgM(41));
          '6':Writeln(fl,GetMsgM(56));
          '8':Writeln(fl,GetMsgM(57));
          'B':Writeln(fl,GetMsgM(42));
          '1':Writeln(fl,GetMsgM(43));
          'I':Writeln(fl,GetMsgM(57));
          'J':Writeln(fl,GetMsgM(50));
          'P':Writeln(fl,GetMsgM(52));
          'L':Writeln(fl,GetMsgM(53));
          'M':Writeln(fl,GetMsgM(54));
          'W':Writeln(fl,GetMsgM(55));
          'H':Writeln(fl,GetMsgM(47));
          'K':Writeln(fl,GetMsgM(51));
          Else
          Dbg('Illegal c2 ('+Inst.c2+') in double parameter instruction!');
          End;
        End;
      End;
    End;
  End;
End;

Procedure Error(ne:Byte;s1,s2:String); { Process error messages }
Begin
If ne>nMaxErr Then Dbg('Illegal error message ('+IntToStr(ne)+')!');
Inc(nerr);
If Not fQuiet Then
  Begin
  If fErrLong Then
    Begin
    Writeln(GetMsgM(3),GetMsgE(ne,s1,s2));
    Writeln('  ',GetMsgM(4),spFile,', ',GetMsgM(5),npLine);
    Writeln('  ',GetMsgM(6),cl.sLine);
    If Not fAnsiOff Then Writeln;
    End Else
    Begin
    Writeln(GetMsgM(3),'==> ',cl.sLine);
    Writeln('[',spFile,',',npLine,'] ',GetMsgE(ne,s1,s2));
    If Not fAnsiOff Then Writeln;
    End;
  End;
If Not fListOff Then
  Begin
  If fErrLong Then
    Begin
    Writeln(fl,' ===> ',GetMsgM(3),GetMsgE(ne,s1,s2));
    Writeln(fl,'      ',GetMsgM(5),cl.sLine);
    Writeln(fl,'      ',GetMsgM(4),spFile,', ',GetMsgM(5),npLine);
    If fBeginner Then BeginnersError;
    End Else
    Begin
    Writeln(fl,GetMsgM(3),'==> ',cl.sLine);
    Writeln(fl,'[',spFile,',',npLine,'] ',GetMsgE(ne,s1,s2));
    End;
  End;
If fErrLong Then
  Begin
  Writeln(fx,' ===> ',GetMsgM(3),GetMsgE(ne,s1,s2));
  Writeln(fx,'      ',GetMsgM(5),cl.sLine);
  Writeln(fx,'      ',GetMsgM(4),spFile,', ',GetMsgM(5),npLine);
  End Else
  Begin
  Writeln(fx,GetMsgM(3),'==> ',cl.sLine);
  Writeln(fx,'[',spFile,',',npLine,'] ',GetMsgE(ne,s1,s2));
  End;
End;

Procedure Warning(f:Boolean;nws:Byte;s1,s2:String); { List warnings }
Begin
Inc(nw);
If Not fQuiet Then
  Begin
  Writeln(GetMsgM(7),GetMsgW(nws,s1,s2));
  If f Then
    Begin
    Writeln('  ',GetMsgM(4),spFile,', ',GetMsgM(5),npLine);
    Writeln('  ',GetMsgM(6),cl.sLine);
    End;
  Writeln;
  End;
If Not fListOff Then
  Begin
  Writeln(fl,' -> ',GetMsgM(7),GetMsgW(nws,s1,s2));
  If f Then
    Begin
    Writeln(fl,'   ',GetMsgM(4),spFile,', ',GetMsgM(5),npLine);
    Writeln(fl,'   ',GetMsgM(6),cl.sLine);
    End;
  End;
End;

Procedure Message(s:String); { Display a message }
Begin
Writeln('  ==> Message: ',s);
If Not fListOff Then
  Begin
  Writeln(fl,' -> Message: ',s);
  Writeln(fl,'   ',GetMsgM(4),spFile,', ',GetMsgM(5),npLine);
  End;
End;

{ ---------------- }

{ Listing routines }
{ ---------------- }

Function LongInt2HexN(n:Byte;i:LongInt):String;
Var h:LongInt;
  b:Byte;
  s:String;
Begin
s:='';
h:=1<<(4*(n-1));
Repeat
  b:=i Div h;
  i:=i-b*h;
  If b>9 Then s:=s+Char(b+55) Else s:=s+Char(b+48);
  h:=h Div 16;
  Until h<=1;
If i>9 Then s:=s+Char(i+55) Else s:=s+Char(i+48);
LongInt2HexN:=s;
End;

Function Longint2Hex(i:LongInt):String;
Var j,k:LongInt;
  s:String;
  f,fs:Boolean;
Begin
j:=268435456;
s:=' ';
f:=False;
If i<0 Then
  Begin
  i:=-i;
  fs:=True;
  End Else
  fs:=False;
Repeat
  k:=i Div j;
  f:=f Or (k>0);
  If f Then
    Begin
    If k>9 Then s:=s+Char(k+55) Else s:=s+Char(k+48);
    End Else
    s:=s+' ';
  i:=i-k*j;
  j:=j Div 16;
  Until j<16;
If i>9 Then s:=s+Char(i+55) Else s:=s+Char(i+48);
If fs Then
  Begin
  k:=1;
  While s[k]=' ' Do Inc(k);
  s[k-1]:='-';
  End;
LongInt2Hex:=s;
End;

Procedure ListSymbols;
Const sListTypes='TLMGEDRCV';
Var k:Byte;
  p:pSList;
Begin
ConcatAllMacroLabels;
Writeln(fl);
If FindAnySymbol Then
  Begin
  Writeln(fl,GetMsgM(8));
  Writeln(fl,GetMsgM(9));
  For k:=1 To Length(sListTypes) Do
    Begin
    p:=FindSymbolFirst(sListTypes[k]);
    While p<>NIL Do With p^ Do
      Begin
      If Not fDefInc Then
        Writeln(fl,'  ',cType,nDefined:6,nUsed:6,iValue:11,Longint2Hex(iValue),' ',sName);
      p:=FindSymbolNext;
      End
    End;
  End Else
  Writeln(fl,GetMsgM(10));
RemoveAllMacroLabels;
End;

{$IFDEF debug}

Procedure DisplaySymbols;
Const sListTypes='TLMGEDRCV';
Var k:Byte;
  p:pSList;
Begin
Writeln;
If FindAnySymbol Then
  Begin
  Writeln(GetMsgM(8));
  Writeln(GetMsgM(9));
  For k:=1 To Length(sListTypes) Do
    Begin
    p:=FindSymbolFirst(sListTypes[k]);
    While p<>NIL Do With p^ Do
      Begin
      If Not fDefInc Then
        Writeln('  ',cType,nDefined:6,nUsed:6,iValue:11,Longint2Hex(iValue),' ',sName);
      p:=FindSymbolNext;
      End
    End;
  End Else
  Writeln(GetMsgM(10));
Writeln;
End;

{$ENDIF}

Procedure CheckSymbols;
Var nUnused,nVarReDef:LongInt;
Begin
If Not CheckAllSymbols(nUnused,nVarReDef) Then
  Begin
  If nUnUsed>0 Then
    Warning(False,1,IntToStr(nUnused),'');
  If nVarReDef>0 Then
    Warning(False,2,'','');
  End;
End;

Function ListGetLine:String;
Begin
Str(npLine:6,ListGetLine);
ListGetLine:=ListGetLine+': ';
End;

Procedure ListOutLine;
Begin
If Not fListOff Then
  Writeln(fl,ListGetLine,cl.sLine);
End;

Function ListGetAddressLongInt:LongInt;
Begin
Case cSegm Of
  'C':ListGetAddressLongInt:=pc;
  'E':ListGetAddressLongInt:=pe;
  'D':ListGetAddressLongInt:=pd;
  End;
End;

Function ListGetLineAndAddress:String;
Var i:LongInt;
Begin
i:=ListGetAddressLongInt;
ListGetLineAndAddress:=ListGetLine+LongInt2HexN(6,i)+' ';
End;

Function ListGetAddressAlone:String;
Var i:LongInt;
Begin
i:=ListGetAddressLongInt;
ListGetAddressAlone:='        '+LongInt2HexN(6,i)+' ';
End;

Procedure ListOutCommand;
Begin
If Not fListOff Then
  Begin
  If fMacroDef Or fIfDef Then
    Writeln(fl,ListGetLine+Copy(cl.sLine,cl.pInst,255)) Else
    Begin
    Writeln(fl,ListGetLineAndAddress,'  ',LongInt2HexN(4,w1),'  ',Copy(cl.sLine,cl.pInst,255));
    If Inst.nW=2 Then
      Writeln(fl,'        ',Longint2HexN(6,ListGetAddressLongInt+1),'   ',LongInt2HexN(4,w2));
    End;
  End;
End;

Procedure ListMacros;
Var se:String;
Begin
If ResetMacroPointer Then
  Begin
  Writeln(fl);
  Writeln(fl,GetMsgM(11));
  Writeln(fl,GetMsgM(12));
  While EnumMacroList(se) Do Writeln(fl,se);
  End Else
  Writeln(fl,GetMsgM(13));
End;

{ -------------------- }
{ Evaluate expressions }
{ -------------------- }
Function CheckSymbolLegal(s:String):Boolean;
Var k:Byte;
  Inst:TInstruction;
Begin
CheckSymbolLegal:=True;
With Inst Do If ((s[1]>='A') And (s[1]<='Z')) Or (s[1]='_') Then
  Begin
  For k:=2 To Length(s) Do
    Case s[k] Of
      'A'..'Z':;
      '0'..'9':;
      '_':;
      Else
      Error(1,s[k],'');
      CheckSymbolLegal:=False;
      End;
  sM:=s;
  If FindInstructionFromMnemonic(Inst) And (Inst.sM<>'OR') And (Inst.sM<>'BRTS') Then
    Begin
    Error(2,s,'');
    CheckSymbolLegal:=False;
    End;
  End Else
  Begin
  Error(3,s,'');
  CheckSymbolLegal:=False;
  End;
End;

Function SnipOff(Var s:String):Boolean;
Begin
While (s<>'') And ((s[1]=' ') Or (s[1]=cTab)) Do Delete(s,1,1);
While (s<>'') And ((s[Length(s)]=' ') Or (s[Length(s)]=cTab)) Do Delete(s,Length(s),1);
SnipOff:=s<>'';
End;

Procedure ExchangeDouble(Var s:String;sr:String;ce:Char);
Var p:Byte;
Begin
p:=Pos(sr,s);
Delete(s,p,2);
Insert(ce,s,p);
End;

Procedure RemoveDoubleAddSub(Var s:String);
Begin
While Pos('--',s)>0 Do ExchangeDouble(s,'--','+');
While Pos('++',s)>0 Do ExchangeDouble(s,'++','+');
While Pos('-+',s)>0 Do ExchangeDouble(s,'-+','-');
While Pos('+-',s)>0 Do ExchangeDouble(s,'+-','-');
End;

Function GetInnerBracket(s:String;Var pka,pke:Byte):Boolean;
Var k,nke,mke:Byte;
  f:Boolean;
Begin
f:=True;
pka:=0;
pke:=0;
nke:=0;
mke:=0;
For k:=1 To Length(s) Do
  Begin
  If f And (s[k]='(') Then
    Begin
    Inc(nke);
    If nke>=mke Then
      Begin
      mke:=nke;
      pka:=k;
      End;
    End;
  If f And (s[k]=')') Then
    Begin
    If nke=mke Then pke:=k;
    Dec(nke);
    If nke=0 Then f:=False;
    End;
  End;
GetInnerBracket:=mke>0;
End;

Function CheckChar(c:Char):Boolean;
Begin
Case c Of
  '!','~','*','/','+','-','<','>','=','&','^','|','(',')':CheckChar:=False;
  Else
  CheckChar:=True;
  End;
End;

Function GetBinaryValue(s:String;Var i:LongInt):Boolean;
Var k:Byte;
Begin
GetBinaryValue:=True;
i:=0;
k:=0;
While GetBinaryValue And (k<Length(s)) Do
  Begin
  Inc(k);
  Case s[k] Of
    '0':i:=i+i;
    '1':i:=i+i+1;
    '.':;
    Else
    GetBinaryValue:=False;
    End;
  End;
If Not GetBinaryValue Then
  Error(4,s[k],'');
End;

Function GetHexValue(s:String;Var i:LongInt):Boolean;
Var k:Byte;
Begin
GetHexValue:=True;
i:=0;
k:=0;
While GetHexValue And (k<Length(s)) Do
  Begin
  Inc(k);
  Case s[k] Of
    '0'..'9':i:=i*16+Byte(s[k])-48;
    'A'..'F':i:=i*16+Byte(s[k])-55;
    '.':;
    Else
    GetHexValue:=False;
    End;
  End;
If Not GetHexValue Then
  Error(5,s[k],'');
End;

Function GetDecimalValue(s:String;Var i:LongInt):Boolean;
Var p:Byte;
Begin
GetDecimalValue:=True;
i:=0;
For p:=1 To Length(s) Do
  Case s[p] Of
    '0'..'9':i:=i*10+(Byte(s[p])-48);
    Else
    GetDecimalValue:=False;
    End;
If Not GetDecimalValue Then
  Error(6,s,'');
End;

Function GetLiteralValue(s:String;Var i:LongInt):Boolean;
Begin
GetLiteralValue:=(Length(s)=3) And (s[3]=s[1]);
If GetLiteralValue Then i:=Byte(s[2]);
End;

Function GetPCvalue(s:String;Var i:LongInt):Boolean;
Begin
If UpperCase(s)='PC' Then
  Begin
  i:=pc+Inst.nw;
  GetPCvalue:=True;
  End Else
  GetPCvalue:=False;
End;

Function GetSingleValue(s:String;Var fValid:Boolean;Var i:LongInt):Boolean;
Var p:Byte;
  fs:Boolean;
Begin
p:=1;
If s[p]='-' Then
  Begin
  fs:=True;
  Inc(p);
  End Else
  fs:=False;
GetSingleValue:=False;
Case s[p] Of
  'A'..'Z','_':
    Begin
    fValid:=True;
    fValid:=GetPCvalue(Copy(s,p,255),i);
    If fValid Then GetSingleValue:=True Else
      Begin
      fValid:=True;
      fValid:=GetSymbolValue('CVLGM',Copy(s,p,255),fValid,i) And fValid;
      If fValid Or (pass=1) Then GetSingleValue:=True Else
        Begin
        fValid:=False;
        Error(7,s,'');
        End;
      End;
    End;
  '0': If Length(s)>2 Then
    Begin
    Case s[p+1] Of
      'B':GetSingleValue:=GetBinaryValue(Copy(s,p+2,255),i);
      'X':GetSingleValue:=GetHexValue(Copy(s,p+2,255),i);
      Else
      GetSingleValue:=GetDecimalValue(Copy(s,p,255),i);
      End;
    End Else
    Begin
    If s='0' Then
      Begin
      i:=0;
      GetSingleValue:=True;
      End Else
      Dbg('GetSingleValue: Don''t understand "'+s+'"!');
    End;
  '1'..'9':GetSingleValue:=GetDecimalValue(Copy(s,p,255),i);
  '$':GetSingleValue:=GetHexValue(Copy(s,p+1,255),i);
  '"','''':GetSingleValue:=GetLiteralValue(s,i);
  End;
If Not GetSingleValue Then
  fValid:=False Else
  Begin
  If fs Then i:=-i;
  End;
End;

Function GetExpressionRight(s:String;Var p:Byte;Var fValid:Boolean;Var i:LongInt):Boolean;
Var sh:String;
Begin
GetExpressionRight:=False;
i:=0;
sh:='';
Inc(p);
If (p<Length(s)) And (s[p]='-') Then
  Begin
  sh:='-';
  Inc(p);
  End;
While (p<=Length(s)) And CheckChar(s[p]) Do
  Begin
  sh:=sh+s[p];
  Inc(p);
  End;
If sh<>'' Then
  GetExpressionRight:=GetSingleValue(sh,fValid,i);
End;

Function GetExpressionLeft(s:String;Var p:Byte; Var fValid:Boolean;Var i:LongInt;fChkSigned:Boolean):Boolean;
Var sh:String;
Begin
GetExpressionLeft:=False;
i:=0;
sh:='';
Dec(p);
While (p>0) And CheckChar(s[p]) Do
  Begin
  sh:=s[p]+sh;
  Dec(p);
  End;
If sh='' Then
  GetExpressionLeft:=True Else
  GetExpressionLeft:=GetSingleValue(sh,fValid,i);
If (p>0) And fChkSigned And (s[p]='-') Then
  Begin
  Dec(p);
  i:=-i;
  End;
End;

Function GetLogicOp(s:String;Var p:Byte;Var bType:Byte):Boolean;
Const
  as:Array[1..9] Of String = ('&&','||','!=','==','>=','<=','<>','>','<');
Begin
bType:=0;
Repeat Inc(bType) Until (bType>9) Or ((Pos(as[bType],s)>0) And ((bType<8) Or (s[Pos(as[bType],s)+1]<>as[bType])));
GetLogicOp:=bType<=9;
If GetLogicOp Then p:=Pos(as[bType],s);
End;

Function GetBitOp(s:String;Var p:Byte;Var bType:Byte):Boolean;
Const
  as:Array[1..5] Of String = ('<<','>>','&','^','|');
Begin
bType:=0;
Repeat Inc(bType) Until (bType>5) Or (Pos(as[bType],s)>0);
GetBitOp:=bType<=5;
If GetBitOp Then p:=Pos(as[bType],s);
End;

Function GetCalcOp(s:String;Var p:Byte;Var bType:Byte):Boolean;
Var p1,p2:Byte;
Begin
GetCalcOp:=False;
p1:=Pos('*',s);
p2:=Pos('/',s);
If (p1<>0) Or (p2<>0) Then { * or / }
  Begin
  p:=p2;
  bType:=1;
  If (p2=0) Or ((p1>0) And (p1<p2)) Then
    Begin
    bType:=2;
    p:=p1;
    End;
  GetCalcOp:=True;
  End Else
  Begin { + or - }
  p1:=Pos('+',s);
  p2:=Pos('-',s);
  If p2=1 Then
    Begin
    p2:=Pos('-',Copy(s,2,255));
    If p2>0 Then p2:=p2+1;
    End;
  If (p1>0) Or (p2>0) Then
    Begin
    p:=p2;
    bType:=3;
    If (p2=0) Or ((p1>0) And (p1<p2)) Then
      Begin
      bType:=4;
      p:=p1;
      End;
    GetCalcOp:=True;
    End;
  End;
End;

Function GetValue(s:String;Var fValid:Boolean;Var i:LongInt):Boolean;
Var bType,pa,p,pe:Byte;
  i1,i2:LongInt;
  c:Char;
Begin
GetValue:=True;
p:=Pos('!',s);
If (p>=Length(s)) Or ((p>1) And (s[p-1]='''') And (s[p+1]='''')) Then p:=0;
While GetValue And (p>0) And (Length(s)>1) And (s[p+1]<>'=') Do
  Begin
  pe:=p;
  GetValue:=GetExpressionRight(Copy(s,p,255),pe,fValid,i);
  If GetValue Then
    Begin
    If i=0 Then i:=1 Else i:=0;
    Delete(s,p,pe-p);
    Insert(IntToStr(i),s,p);
    p:=Pos('!',s);
    End;
  End;
While GetValue And (Pos('''',s)>0) Do
  Begin
  p:=Pos('''',s);
  GetValue:=Length(s)>=(p+2);
  If GetValue Then
    Begin
    Case s[p+1] Of
      '''':Begin
        GetValue:=(Length(s)>(p+3)) And (s[p+2]='''') And (s[p+3]='''');
        If GetValue Then
          Begin
          Delete(s,p,4);
          Insert(IntToStr(Byte('''')),s,p);
          End;
        End;
      '\':Begin
        GetValue:=(Length(s)>(p+3)) And (s[p+3]='''');
        If GetValue Then
          Begin
          c:=s[p+2];
          Delete(s,p,4);
          If s[p+2]='\' Then
            Insert(IntToStr(Byte('\')),s,p) Else
            Insert(IntToStr(Byte(s[p+2]) Mod $20),s,p);
          End;
        End;
      Else
      c:=s[p+1];
      Delete(s,p,3);
      Insert(IntToStr(Byte(c)),s,p);
      End;
    End;
  End;
While GetValue And GetLogicOp(s,p,bType) Do
  Begin
  pa:=p;
  If bType<8 Then pe:=p+1 Else pe:=p;
  GetValue:=GetExpressionLeft(s,pa,fValid,i1,False) And GetExpressionRight(s,pe,fValid,i2);
  If GetValue Then
    Begin
    i:=0;
    Case bType Of
      1:If (i1<>0) And (i2<>0) Then i:=1; {&&}
      2:If (i1<>0) Or (i2<>0) Then i:=1; {||}
      3:If ((i1<>0) And (i2=0)) Or ((i1=0) And (i2<>0)) Then i:=1; {!=}
      4:If i1=i2 Then i:=1; {==}
      5:If i1>=i2 Then i:=1; {>=}
      6:If i1<=i2 Then i:=1; {<=}
      7:If i1<>i2 Then i:=1; {<>}
      8:If i1>i2 Then i:=1; {>}
      9:If i1<i2 Then i:=1; {<}
      End;
    Delete(s,pa+1,pe-pa-1);
    Insert(IntToStr(i),s,pa+1);
    End;
  End;
If GetValue And (Pos('=',s)>0) Then
  Begin
  GetValue:=False;
  Error(8,'','');
  End;
p:=Pos('~',s);
While GetValue And (p>0) And (Length(s)>1) Do
  Begin
  pe:=p;
  GetValue:=GetExpressionRight(Copy(s,p,255),pe,fValid,i);
  If GetValue Then
    Begin
    i:=Not i;
    Delete(s,p,pe-p);
    Insert(IntToStr(i),s,p);
    p:=Pos('~',s);
    End;
  End;
While GetValue And GetBitOp(s,p,bType) Do
  Begin
  pa:=p;
  pe:=p;
  If bType<3 Then pe:=p+1;
  GetValue:=GetExpressionLeft(s,pa,fValid,i1,False) And GetExpressionRight(s,pe,fValid,i2);
  If GetValue Then
    Begin
    i:=0;
    Case bType Of
      1:If (Extended(i1)*(2**i2))<=Extended(iMaxInt) Then i:=i1 SHL i2 Else {<<}
        Error(9,IntToStr(i1),IntToStr(i2));
      2:i:=i1 SHR i2; {>>}
      3:i:=i1 AND i2; {&}
      4:i:=i1 XOR i2; {^}
      5:i:=i1 OR i2; {|}
      End;
    Delete(s,pa+1,pe-pa-1);
    Insert(IntToStr(i),s,pa+1);
    End;
  End;
While GetValue And GetCalcOp(s,p,bType) Do
  Begin
  pa:=p;
  pe:=p;
  GetValue:=GetExpressionLeft(s,pa,fValid,i1,True) And GetExpressionRight(s,pe,fValid,i2);
  If Getvalue Then
    Begin
    i:=0;
    Case bType Of
      1:If i2=0 Then
          Begin
          If pass=2 Then Error(101,'','') Else i:=0;
          End Else
          i:=i1 Div i2; {/}
      2:If (Extended(i1)*Extended(i2))<Extended(iMaxInt) Then {*}
          i:=i1*i2 Else
          Error(10,IntToStr(i1),IntToStr(i2));
      3:If (Extended(i1)-Extended(i2))>-Extended(iMaxInt) Then {-}
          i:=i1-i2 Else
          Error(12,IntToStr(i1),IntToStr(i2));
      4:If Extended(i1)+Extended(i2)<Extended(iMaxInt) Then {+}
          i:=i1+i2 Else
          Error(11,IntToStr(i1),IntToStr(i2));
      Else
      Dbg('Illegal bType '+IntToStr(bType)+' in GetValue!');
      End;
    Delete(s,pa+1,pe-pa-1);
    Insert(IntToStr(i),s,pa+1);
    RemoveDoubleAddSub(s);
    End;
  End;
If GetValue Then GetValue:=GetSingleValue(s,fValid,i);
End;

Function GetFunction(s:String;Var pka:Byte;Var i:LongInt):Boolean;
Const
  nas=11;
  as:Array[1..nas] Of String = ('LOW','HIGH','BYTE1','BYTE2','BYTE3','BYTE4',
    'LWRD','HWRD','PAGE','EXP2','LOG2');
Var k,p:Byte;
  sh:String;
Begin
GetFunction:=True;
If pka>3 Then
  Begin
  p:=pka-1;
  sh:='';
  While (p>0) And (CheckChar(s[p])) Do
    Begin
    sh:=s[p]+sh;
    Dec(p)
    End;
  If sh<>'' Then
    Begin
    k:=0;
    Repeat Inc(k) Until (k>nas) Or (sh=as[k]);
    If k<=nas Then
      Begin
      If i<0 Then
        Begin
        i:=-i;
        If i<256 Then i:=256-i Else
        If i<65536 Then i:=65536-i Else
        If i<16777216 Then i:=16777216-i;
	End;
      Case k Of
        1,3:i:=i MOD 256;
        2:i:=i DIV 256;
        4:i:=(i DIV 256) MOD 256;
        5:i:=(i DIV 65536) MOD 256;
        6:i:=(i DIV 16777216) MOD 256;
        7:i:=i MOD 65536;
        8:i:=(i DIV 65536) MOD 65536;
        9:i:=(i DIV 65536) MOD 64;
        10:i:=2 SHL (i-1);
        11:i:=TRUNC(LN(i)/LN(2.0));
        End;
      pka:=p+1;
      End Else
      Begin
      GetFunction:=False;
      Error(13,Copy(s,p+1,pka-p),'');
      End;
    End;
  End;
End;

Function RemoveBlankTab(Var s:String):Boolean;
Var p:Byte;
Begin
While Pos(''' ''',s)>0 Do
  Begin
  p:=Pos(''' ''',s);
  Delete(s,p,3);
  Insert('32',s,p);
  End;
While Pos(' ',s)>0 Do Delete(s,Pos(' ',s),1);
While Pos(cTab,s)>0 Do Delete(s,Pos(cTab,s),1);
RemoveBlankTab:=s<>'';
End;

Function GetExpression(s:String;Var fValid:Boolean;Var i:LongInt):Boolean;
Var pka,pke:Byte;
Begin
fValid:=True;
GetExpression:=True;
RemoveBlankTab(s);
RemoveDoubleAddSub(s);
While GetExpression And GetInnerBracket(s,pka,pke) Do
  Begin
  GetExpression:=GetValue(Copy(s,pka+1,pke-pka-1),fValid,i);
  If GetExpression Then
    Begin
    GetExpression:=GetFunction(s,pka,i);
    If GetExpression Then
      Begin
      Delete(s,pka,pke-pka+1);
      Insert(IntToStr(i),s,pka);
      RemoveDoubleAddSub(s);
      End;
    End;
  End;
GetExpression:=GetExpression And GetValue(s,fValid,i);
End;

Function CheckExpression(s:String):Boolean;
Var k:Byte;
  nk:LongInt;
Begin
CheckExpression:=True;
nk:=0;
For k:=1 To Length(s) Do
  Case s[k] Of
    '(':If nk>=0 Then Inc(nk);
    ')':Dec(nk);
    'A'..'Z':;
    '0'..'9':;
    ' ',cTab:;
    '!','~','*','/','+','-','<','>','=','&','^','|':;
    '$','"','_','''':;
    Else
    Error(14,s[k],'');
    CheckExpression:=False;
    End;
If nk<>0 Then
  Begin
  CheckExpression:=False;
  If nk<0 Then
    Error(15,'','') Else
    Error(16,'','');
  End;
End;

Function GetConstantLongInt(s:String;Var fValid:Boolean;Var i:LongInt):Boolean;
Begin
GetConstantLongInt:=GetExpression(s,fValid,i);
End;

{ ------------------------------------- }
{ Processing parameters of instructions }
{ ------------------------------------- }
Function GetRegister(s:String;Var wr:Word):Boolean;
Var e:Word;
  fValid:Boolean;
  i:LongInt;
Begin
wr:=0;
i:=0;
GetRegister:=False;
If SnipOff(s) Then
  Begin
  If (Length(s)>1) And (s[1]='R') Then
    Begin
    Val(Copy(s,2,255),wr,e);
    If (e=0) And (wr<32) Then GetRegister:=True;
    End;
  If Not GetRegister Then
    Begin
    fValid:=True;
    If GetSymbolValue('R',s,fValid,i) Then
      Begin
      If (i>=0) And (i<32) And fValid Then
        Begin
        wr:=i;
        GetRegister:=True;
        End Else
        Error(17,IntToStr(i),'0..31');
      End Else
      Begin
      If pass=1 Then
        Begin
        wr:=0;
        GetRegister:=True;
        End Else
        Error(18,'','');
      End;
    End;
  End Else
  Error(19,'','');
End;

Function GetHighRegister(s:String;Var wr:Word):Boolean;
Begin
wr:=0;
GetHighRegister:=GetRegister(s,wr);
If GetHighRegister Then
  Begin
  If wr>=16 Then
    wr:=wr-16 Else
    Begin
    If pass=2 Then
      Begin
      GetHighRegister:=False;
      Error(17,IntToStr(wr),'R16..R31');
      End;
    End;
  End;
End;

Function GetDoubleRegister(s:String;Var wr:Word):Boolean;
Begin
wr:=0;
GetDoubleRegister:=GetRegister(s,wr);
If GetDoubleRegister Then
  Begin
  Case wr Of
    24,26,28,30:wr:=(wr-24) Div 2;
    Else
    GetDoubleRegister:=False;
    Error(17,IntToStr(wr),'R24/26/28/30!');
    End;
  End;
End;

Function GetMovwReg(s:String;Var wr:Word):Boolean;
Begin
wr:=0;
GetMovwReg:=GetRegister(s,wr);
If GetMovwReg Then
  Begin
  If ((wr<0) Or (wr>30) Or Odd(wr)) Then
    Begin
    GetMovwReg:=False;
    Error(17,IntToStr(wr),GetMsgM(39));
    End Else
    wr:=wr Div 2;
  End;
End;

Function GetMulReg(s:String;Var wr:Word):Boolean;
Begin
wr:=0;
GetMulReg:=GetRegister(s,wr);
If GetMulReg Then
  Begin
  If (wr>=16) And (wr<24) Then wr:=wr-16 Else
    Error(17,IntToStr(wr),'R16..R23');
  End;
End;

Function GetPort(s:String;Var wr:Word):Boolean;
Var i:LongInt;
  fValid:Boolean;
Begin
wr:=0;
i:=0;
fValid:=True;
GetPort:=GetExpression(s,fValid,i);
If GetPort Then
  Begin
  If (i>=0) And (i<64) Then
    Begin
    wr:=i MOD 64;
    If (pass=2) And (Not fValid) Then
      Begin
      GetPort:=False;
      Error(20,'','');
      End;
    End Else
    Begin
    GetPort:=False;
    Error(21,IntToStr(i),'0..63');
    End;
  End;
End;

Function GetLowPort(s:String;Var wr:Word):Boolean;
Begin
wr:=0;
GetLowPort:=GetPort(s,wr);
If GetLowPort And (wr>=32) Then
  Begin
  GetLowPort:=False;
  Error(21,IntToStr(wr),'0..31');
  End;
End;

Function GetBit(s:String;Var wr:Word):Boolean;
Var fValid:Boolean;
  i:LongInt;
Begin
wr:=0;
i:=0;
fValid:=True;
GetBit:=GetExpression(s,fValid,i) And fValid;
If (Not fValid) And (pass=1) Then
  Begin
  GetBit:=True;
  wr:=0;
  End;
If (pass=2) And GetBit Then
  Begin
  If (i>=0) And (i<8) Then
    wr:=i MOD 8 Else
    Begin
    GetBit:=False;
    Error(22,IntToStr(i),'');
    End;
  End;
End;

Function GetConstant(s:String;Var i:LongInt;imin,imax:LongInt):Boolean;
Var fValid:Boolean;
Begin
fValid:=True;
GetConstant:=GetExpression(s,fValid,i);
If (Not fValid) And (pass=1) Then i:=0;
If GetConstant Then
  Begin
  If (i<imin) Or (i>imax) Then
    Begin
    GetConstant:=False;
    Error(24,IntToStr(i),IntToStr(imin)+'...'+IntToStr(imax));
    End;
  End Else
  Error(25,s,'');
End;

Function GetConstantSmall(s:String;Var wr:Word):Boolean;
Var i:LongInt;
Begin
wr:=0;
GetConstantSmall:=GetConstant(s,i,0,63);
If GetConstantSmall Then wr:=i;
End;

Function GetConstantByte(s:String;Var wr:Word):Boolean;
Var i:LongInt;
Begin
wr:=0;
GetConstantByte:=GetConstant(s,i,-255,255);
If GetConstantByte Then
  Begin
  If i<0 Then i:=256+i;
  wr:=i;
  End;
End;

Function GetConstantByteInverted(s:String; Var wr:Word):Boolean;
Begin
GetConstantByteInverted:=GetConstantByte(s,wr);
wr:=255-wr;
End;

Function GetConstantWord(s:String;Var wr:Word):Boolean;
Var i:LongInt;
Begin
wr:=0;
GetConstantWord:=GetConstant(s,i,-65535,65535);
If GetConstantWord Then
  Begin
  If i<0 Then i:=65536+i;
  wr:=i;
  End;
End;

Function GetConstantLongWord(s:String;Var wr:LongWord):Boolean;
Var fValid:Boolean;
  i:LongInt;
Begin
wr:=0;
fValid:=True;
GetConstantLongWord:=GetExpression(s,fValid,i);
If (Not fValid) And (pass=1) Then i:=0;
If GetConstantLongWord Then
  Begin
  If (pass=2) And (Not fValid) Then
    Begin
    GetConstantLongWord:=False;
    Error(26,'','');
    End Else
    Begin
    If (i>=0) And (i<=4194303) Then
      wr:=i Else
      Begin
      GetConstantLongWord:=False;
      Error(24,IntToStr(i),'0..4194303');
      End;
    End;
  End;
End;

Function GetOffset(s:String;Var wr:Word;imax:LongInt):Boolean;
Var fValid,fError:Boolean;
  i:LongInt;
Begin
wr:=0;
i:=0;
GetOffset:=True;
If pass=1 Then wr:=0 Else
  Begin
  fValid:=True;
  GetOffset:=GetExpression(s,fValid,i);
  If (s[1]='+') Or (s[1]='-') Then
    Begin
    GetOffset:=GetExpression(Copy(s,2,255),fValid,i);
    If s[1]='-' Then i:=-i;
    End Else
    Begin
    GetOffset:=GetExpression(s,fValid,i);
    i:=i-(pc+1);
    End;
  If GetOffset And fValid Then
    Begin
    fError:=False;
    If (i<-imax) Or (i>(imax-1)) Then
      Begin
      fError:=True;
      i:=i+pc+1;
      If fWrapOn Then
        Begin
        If i>(pc+1) Then
          Begin
          i:=i-(pc+1+device.nc Div 2);
          End Else
          Begin
          i:=device.nc Div 2-(pc+1)+i;
          End;
        fError:=(i<-imax) Or (i>(imax-1));
        If Not fError Then Warning(True,7,'','');
        End;
      If fError Then
        Begin
        GetOffset:=False;
        Error(23,IntToStr(i),'-'+IntToStr(imax)+'..+'+IntToStr(imax-1));
        End Else
      End;
    End;
  If GetOffset And fValid Then
    Begin
    If i>=0 Then wr:=i MOD imax Else wr:=2*imax+i;
    End;
  End;
End;

Function GetLdSt(s:String;Var wr:Word;cST:Boolean):Boolean;
Var se:String;
Begin
If cST Then se:='ST' Else se:='LD';
wr:=0;
GetLdSt:=True;
If SnipOff(s) Then
  Begin
  Case Length(s) Of
    1:Begin
      Case s[1] Of
        'X':wr:=$900C;
        'Y':wr:=$8008;
        'Z':wr:=$8000;
        Else
        GetLdSt:=False;
        Error(27,se,s);
        End;
      End;
    2:If s[1]='-' Then
      Begin
      If CheckInstValid(hasLdXY) Then
        Begin
        Case s[2] Of
          'X':wr:=$900E;
          'Y':wr:=$900A;
          'Z':wr:=$9002;
          Else
          GetLdSt:=False;
          Error(27,se,s);
          End;
        End Else
        Begin
        Error(98,'','');
        GetLdSt:=False;
        End;
      End Else
      Begin
      If s[2]='+' Then
        Begin
        If CheckInstValid(hasLdXY) Then
          Begin
          Case s[1] Of
            'X':wr:=$900D;
            'Y':wr:=$9009;
            'Z':wr:=$9001;
            Else
            GetLdSt:=False;
            Error(27,se,s);
            End;
          End Else
          Begin
          Error(98,'','');
          GetLdSt:=False;
          End;
        End Else
        Begin
        GetLdSt:=False;
        Error(27,se,s);
        End;
      End;
    Else
    GetLdSt:=False;
    Error(27,se,s);
    End;
  If GetLdSt And cST Then wr:=wr+$0200;
  End Else
  Begin
  GetLdSt:=False;
  Error(28,se,'');
  End;
End;

Function GetLddStd(s:String;Var wr:Word;fStd:Boolean):Boolean;
Var se:String;
  wh:Word;
  fValid:Boolean;
  i:LongInt;
Begin
If fStd Then se:='STD' Else se:='LDD';
wr:=0;
GetLddStd:=True;
If RemoveBlankTab(s) Then
  Begin
  Case s[1] Of
    'Y':wr:=$8008;
    'Z':wr:=$8000;
    Else
    GetLddStd:=False;
    Error(29,se,s[1]);
    End;
  If GetLddStd And (Length(s)>1) And (s[2]='+') Then
    Begin
    GetLddStd:=GetExpression(Copy(s,3,255),fValid,i);
    If (Not fValid) And (pass=1) Then i:=0;
    If GetLddStd Then
      Begin
      If (i>=0) And (i<64) Then
        Begin
        wh:=i MOD 64;
        wr:=wr+(wh AND $07)+128*(wh AND $18)+256*(wh AND $20);
        If fStd Then wr:=wr+$0200;
        End Else
        Begin
        GetLddStd:=False;
        Error(30,se,IntToStr(i));
        End;
      End;
    End;
  End Else
  Begin
  GetLddStd:=False;
  Error(31,'','');
  End;
End;

Function GetELpm(s1,s2:String;Var w1,w2:Word):Boolean;
Begin
GetELpm:=False;
w1:=0;
w2:=0;
If RemoveBlankTab(s1) Or RemoveBlankTab(s2) Then
  Begin
  GetELpm:=GetRegister(s1,w2);
  If GetELpm Then
    Begin
    If s2<>'' Then
      Begin
      If s2[1]='Z' Then
        Begin
        GetELpm:=True;
        Case Length(s2) Of
          1:If Inst.sM='ELPM' Then
              Begin
              w1:=$0006;
              If Not CheckInstValid(hasElpmZ) Then
                Error(35,'ELPM r,Z',Device.sn);
              End Else
              Begin
              w1:=$0004;
              If Not CheckInstValid(hasLpmZ) Then
                Error(35,'LPM r,Z',Device.sn);
              End;
          2:If s2[2]='+' Then
            Begin
            If Inst.sM='ELPM' Then
              Begin
              w1:=$0007;
              If Not CheckInstValid(hasElpmZ) Then
                Error(35,'ELPM r,Z+',Device.sn);
              End Else
              Begin
              w1:=$0005;
              If Not CheckInstValid(hasLpmZ) Then
                Error(35,'LPM r,Z+',Device.sn);
              End;
            End Else
            Begin
            Error(32,s2[2],'');
            GetELpm:=False;
            End;
          Else
          Error(33,s2,'');
          GetELpm:=False;
          End;
        End Else
        Error(33,s2,'');
      End Else
      Error(33,'-','');
    End Else
    Error(34,'','');
  End Else
  Begin
  GetELpm:=True;
  If Inst.sM='ELPM' Then
    Begin
    w1:=$05D8;
    If Not CheckInstValid(hasElpm) Then
      Error(35,'ELPM',Device.sn);
    End Else
    Begin
    w1:=$05C8;
    If Not CheckInstValid(hasLpm) Then
      Error(35,'LPM',Device.sn);
    End;
  End;
End;

{ -------------------- }
{ Analyse instructions }
{ -------------------- }
Function AnalyseCommand:Boolean;
Var wp1,wp2:Word;
  lw1:LongWord;
Begin
AnalyseCommand:=False;
With Inst Do With cl Do
  Begin
  If Not CheckInstValid(wt) Then
    Error(35,sM,device.sn) Else
    Begin
    If (np<0) Or (np=nsp) Then
      Begin
      w1:=wx;
      Case c1 Of
        ' ':Begin
          AnalyseCommand:=(nsp=0);
          If nsp>0 Then Error(90,'','');
          End;
        'R':AnalyseCommand:=GetRegister(asp[1],wp1);
        'H':AnalyseCommand:=GetHighRegister(asp[1],wp1);
        'D':AnalyseCommand:=GetDoubleRegister(asp[1],wp1);
        'P':AnalyseCommand:=GetPort(asp[1],wp1);
        'Q':AnalyseCommand:=GetLowPort(asp[1],wp1);
        'B':AnalyseCommand:=GetBit(asp[1],wp1);
        '1':AnalyseCommand:=GetOffset(asp[1],wp1,64);
        '2':AnalyseCommand:=GetOffset(asp[1],wp1,2048);
        'S':AnalyseCommand:=GetLdSt(asp[1],wp1,True);
        'T':AnalyseCommand:=GetLddStd(asp[1],wp1,True);
        'W':AnalyseCommand:=GetConstantWord(asp[1],wp1);
        'X':AnalyseCommand:=GetConstantLongWord(asp[1],lw1);
        'E':AnalyseCommand:=GetELpm(asp[1],asp[2],wp1,wp2); { LPM,ELPM }
        'J':AnalyseCommand:=GetMulReg(asp[1],wp1);
        'K':AnalyseCommand:=GetMovwReg(asp[1],wp1);
        Else
        Dbg('Illegal parameter type c1 ('+c1+')!');
        End;
      Case c2 Of
        ' ':Begin
          AnalyseCommand:=(nsp<=1) Or (sM='LPM') Or (sM='ELPM');
          If Not AnalyseCommand Then
            Error(90,'','');
          End;
        'R':AnalyseCommand:=GetRegister(asp[2],wp2);
        'H':AnalyseCommand:=GetHighRegister(asp[2],wp2);
        'I':AnalyseCommand:=GetConstantByteInverted(asp[2],wp2);
        'P':AnalyseCommand:=GetPort(asp[2],wp2);
        'B':AnalyseCommand:=GetBit(asp[2],wp2);
        '1':AnalyseCommand:=GetOffset(asp[2],wp2,64);
        '6':AnalyseCommand:=GetConstantSmall(asp[2],wp2);
        '8':AnalyseCommand:=GetConstantByte(asp[2],wp2);
        'L':AnalyseCommand:=GetLdSt(asp[2],wp2,False);
        'M':AnalyseCommand:=GetLddStd(asp[2],wp2,False);
        'W':AnalyseCommand:=GetConstantWord(asp[2],wp2);
        'J':AnalyseCommand:=GetMulReg(asp[2],wp2);
        'K':AnalyseCommand:=GetMovwReg(asp[2],wp2);
        Else
        Dbg('Illegal parameter type c2 ('+c2+')!');
        End;
      End Else
      Begin
      If nsp>np Then Error(90,'','') Else Error(91,'','');
      AnalyseCommand:=False;
      End;
    If AnalyseCommand Then
      Begin
      If (c1<>' ') And (c2<>' ') Then
        Case cC Of
          '1':w1:=w1+16*wp1+(wp2 MOD 16)+32*(wp2 AND $10); { ADC,ADD,CP,CPC,CPSE,EOR,MOV,OR,SBC,SUB }
          '2':w1:=w1+16*wp1+(wp2 MOD 16)+16*(wp2 AND $F0); { ANDI,CBR,CPI,LDI,ORI,SBCI,SBR,SUBI }
          '3':w1:=w1+16*wp1+wp2; { BLD,BST,SBRC,SBRS }
          '4':w1:=w1+wp1+8*wp2; { BRBC,BRBS }
          '5':w1:=w1+8*wp1+wp2; { CBI,SBI,SBIC,SBIS }
          '6':w1:=w1+16*wp1+(wp2 MOD 16)+32*(wp2 AND $30); { IN }
          '7':w1:=w1+16*wp1+(wp2 MOD 16)+4*(wp2 AND $30); { SBIW,ADIW }
          '8':w1:=wp2+16*wp1; { LDD }
          '9': {LDS}
            Begin
            w1:=w1+16*wp1;
            w2:=wp2;
            End;
          'A': {STS}
            Begin
            w1:=w1+16*wp2;
            w2:=wp1;
            End;
          'B':w1:=wp2+16*wp1; { LD }
          'C':w1:=wp1+16*wp2; { ST }
          'D':w1:=wp1+16*wp2; { STD }
          'E':w1:=w1+16*wp2+(wp1 MOD 16)+32*(wp1 AND $30); { OUT }
          'F':w1:=w1+16*wp1+wp2; { FMUL,FMULS,FMULSU,MOVW,MULS,MULSU }
          Else
          Dbg('Illegal 2-parameter method '+cC+'!');
          End;
      If (c1<>' ') And (c2=' ') Then
        Case cC Of
          '1':w1:=w1+16*(wp1 MOD 32); { ASR,COM,DEC,INC,LSR,NEG,POP,PUSH,ROR,SWAP }
          '2':w1:=w1+16*(wp1 MOD 8); { BCLR,BSET }
          '3':w1:=w1+8*(wp1 MOD 128); { BRxx }
          '4':w1:=w1+16*wp1+(wp1 MOD 16)+32*(wp1 AND $10); { CLR,LSL,ROL,TST }
          '5':w1:=w1+wp1; { RCALL,RJMP }
          '6': Begin { CALL,JMP }
            w2:=lw1 AND $FFFF;
            wp1:=lw1 DIV $010000;
            w1:=w1+(lw1 AND $0001)+8*(lw1 AND $003E);
            End;
          '7':w1:=w1+16*wp1; { SER }
          '8':w1:=w1+wp1+16*wp2; { LPM,ELPM }
          Else
          Dbg('Illegal 1-parameter method '+cC+'!');
          End;
      End Else
      Begin
      w1:=0;
      w2:=0;
      End;
    End;
  End;
End;

{ ---------------------------- }
{ Process assembler directives }
{ ---------------------------- }

Procedure CheckDeviceError(i:Integer;sn:String);
Begin
Case i Of
  -1:Error(92,sn,'');
  -2:Error(93,'','');
  -3:Error(94,'','');
  -4:Error(95,'','');
  -5:Error(96,'','');
  -6:Error(100,'','');
  -7:Error(97,'','');
  End;
End;

Procedure ProcessDevice;
Begin
With device Do
  Begin
  sn:=cl.asp[1];
  If (sn[1]='"') And (sn[Length(sn)]='"') Then
    Begin
    Delete(sn,Length(sn),1);
    Delete(sn,1,1);
    End;
  If GetAvrType Then
    Begin
    If pass=1 Then CheckDeviceError(DefDevice(sn),sn); { calling gavrdefs }
    fWrapOn:=fWrap;
    End Else
    Error(53,'','');
  End;
End;

Procedure IncludeFile;
Var s,spFileOld:String;
  npLineOld:LongInt;
  i:Integer;
Begin
s:=cl.asp[1];
If SnipOff(s) Then
  Begin
  If s[1]='"' Then Delete(s,1,1);
  If s[Length(s)]='"' Then Delete(s,Length(s),1);
  If Pos('DEF.INC',Uppercase(s))>0 Then
    Begin
    If pass=1 Then i:=InclDevice(s) Else i:=0;
    If i=0 Then Warning(True,9,'','');
    CheckDeviceError(i,'def.inc');
    End Else
    Begin
    If FileExists(s) Then
      Begin
      spFileOld:=spFile;
      npLineOld:=npLine;
      Writeln(GetMsgM(14),s);
      If Not fListOff Then Writeln(fl,'  ',GetMsgM(14),s);
      ProcessFile(s);
      Writeln('ok.');
      spFile:=spFileOld;
      npLine:=npLineOld;
      If Not fListOff Then Writeln(fl,'  ',GetMsgM(15),spFile);
      End Else
      Begin
      Error(36,s,'');
      End;
    End;
  End Else
  Error(37,'','');
End;

Function SplitPseudoEquation(Var s1,s2:String):Boolean;
Var p:Byte;
Begin
SplitPseudoEquation:=False;
p:=Pos('=',cl.asp[1]);
If p>0 Then
  Begin
  s1:=Copy(cl.asp[1],1,p-1);
  s2:=Copy(cl.asp[1],p+1,255);
  If SnipOff(s1) Then
    Begin
    If SnipOff(s2) Then
      Begin
      If CheckSymbolLegal(s1) Then
        Begin
        If CheckExpression(s2) Then
          SplitPseudoEquation:=True Else
          Error(38,'2',s2);
        End Else
        Error(38,'1',s1);
      End Else
      Error(38,'2',s2);
    End Else
    Error(38,'1',s1);
  End Else
  Error(39,'','');
End;

Procedure EquSetDef(cType:Char);
Var s1,s2:String;
  w:Word;
  i,ig:LongInt;
  fv,fValid:Boolean;
Begin
If SplitPseudoEquation(s1,s2) Then
  Begin
  Case cType Of
    'R':
      Begin
      fvalid:=True;
      If GetSymbolValue('CVG',s1,fValid,i) Then
        Begin
        fv:=False;
        Error(40,s1,'constant or variable');
        End Else
        Begin
        fv:=GetRegister(s2,w);
        i:=w;
        End;
      End;
    'C','V':
      Begin
      fValid:=True;
      If GetSymbolValue('R',s1,fValid,i) Then
        Begin
        fv:=False;
        Error(40,s1,'register');
        End Else
	fv:=GetConstantLongInt(s2,fValid,i);
      End;
    Else
    Dbg('EquSetDef: cType not R, C, G or V!');
    End;
{  If fHasMacLines Then
    Begin
    Case cType Of
      'R','C':Error(55,s1,'');
      'V':Begin
          If GetSymbolValue('GVM',s1,FALSE,ig) Then
            SetSymbol('GVM',s1,fValid,i) Else
	    Dbg('EquSetDef: Variable missing!');
	  End;
      End;
    End Else
    Begin}
    If fv Then
      Begin
      Case pass Of
        1:If Not AddSymbol(cType,s1,fValid,i) Then Error(55,s1,'');
        2:If Not SetSymbol(cType,s1,fValid,i) Then Error(56,s1,'');
        End;
      End Else
      Error(41,'','');
{    End;}
{$IFDEF debug}
    DisplaySymbols;
{$ENDIF}
  End;
End;

Procedure ProcessSetGlobal;
Var k:Byte;
  fValid:Boolean;
  i:LongInt;
Begin
{$IFDEF debug}
DisplaySymbols;
{$ENDIF}
For k:=1 To cl.nsp Do
  Begin
  fValid:=False;
  If (Not GetSymbolValue('G',UpperCase(cl.asp[k]),fValid,i)) And
    (Not ExportMacroSymbol(Uppercase(cl.asp[k]))) Then Error(91,'','');
  End;
End;

Procedure ProcessByte;
Var i:LongInt;
  fValid:Boolean;
Begin
If GetExpression(cl.asp[1],fValid,i) And fValid Then
  Begin
  pd:=pd+i;
  nd:=nd+i;
  End Else
  Begin
  If pass=2 Then Error(43,'','');
  End;
End;

Procedure ProcessOrg;
Var i:LongInt;
  fValid:Boolean;
Begin
If GetExpression(cl.asp[1],fValid,i) And fValid Then
  Begin
  Case cSegm Of
    'C':If i>=pc Then pc:=i Else
      Error(46,IntToStr(i),'code segment');
    'D':If i>=pd Then pd:=i Else
      Error(46,IntToStr(i),'data segment');
    'E':If i>=pe Then pe:=i Else
      Error(46,IntToStr(i),'eeprom segment!');
    End;
  End Else
  Begin
  If pass=2 Then Error(47,'','');
  End;
End;

Function GetString(s:String):String;
Var k:Byte;
  sh:String;
Begin
sh:=s;
GetString:='';
If (s[1]='"') And (s[Length(s)]='"') Then
  Begin
  Delete(s,1,1);
  Delete(s,Length(s),1);
  k:=Length(s);
  While k>0 Do
    Begin
    Case s[k] Of
      '"':If (k>1) And (s[k-1]='"') Then
        Begin
        Delete(s,k,1);
        Dec(k);
        End;
      '\':If (k>1) And (s[k-1]='\') Then
        Begin
        Delete(s,k,1);
        Dec(k);
        End Else
        Begin
        Delete(s,k,1);
        If k<=Length(s) Then
          s[k]:=Char(Byte(s[k]) MOD $20) Else
          Error(88,sh,'');
        End;
      End;
    Dec(k);
    End;
  GetString:=s;
  End Else
  Error(89,s,'');
End;

Function GetLiteral(s:String):Char;
Begin
GetLiteral:=Char(0);
If (s[1]='''') And (s[Length(s)]='''') Then
  Begin
  Case Length(s) Of
    3:GetLiteral:=s[2];
    4:Begin
      Case s[2] Of
        '''':If s[3]='''' Then
          GetLiteral:='''' Else
          Error(87,s,'');
        '\':If s[3]='\' Then
          GetLiteral:='\' Else
          GetLiteral:=Char(Byte(s[3]) Mod $20);
        Else
        Error(87,s,'');
        End;
      End;
    Else
    Error(87,s,'');
    End;
  End Else
  Dbg('Literal ('+s+') does not start and end with a ''!');
End;

Procedure ProcessDb;
Var j,k,nab:Byte;
  ab:Array[1..255] Of Byte;
  w:Word;
  s:String;
Begin
nab:=0;
For k:=1 To cl.nsp Do With cl Do
  Case asp[k,1] Of
    '''':Begin
      Inc(nab);
      ab[nab]:=Byte(GetLiteral(asp[k]));
      End;
    '"':Begin
      s:=GetString(asp[k]);
      If s<>'' Then For j:=1 To Length(s) Do
        Begin
        Inc(nab);
        ab[nab]:=Byte(s[j]);
        End;
      End;
    Else
    GetConstantByte(cl.asp[k],w);
    Inc(nab);
    ab[nab]:=w;
    End;
If nab=0 Then
  Warning(True,3,'','') Else
  Begin
  If (cSegm='C') And ((nab MOD 2)=1) Then
    Begin
    If pass=2 Then
      Warning(True,4,'','');
    Inc(nab);
    ab[nab]:=0;
    End;
  s:=ListGetAddressAlone;
  Case cSegm Of
    'C':For k:=1 To nab Div 2 Do
      Begin
      w:=256*ab[2*k]+ab[2*k-1];
      Inc(nconstants);
      If pass=2 Then HexWriteC(pc,w);
      If Not fListOff Then
        Begin
        If (k Mod 4)<>1 Then s:=s+' ';
        s:=s+LongInt2HexN(4,w);
        If ((k MOD 4)=0) Or (k=(nab Div 2)) Then
          Begin
          Writeln(fl,s);
          Inc(pc);
          s:=ListGetAddressAlone;
          Dec(pc);
          End;
        End;
      Inc(pc);
      End;
    'E':For k:=1 To nab Do
      Begin
      If pass=2 Then HexWriteE(pe,ab[k]);
      Inc(ne);
      If Not fListOff Then
        Begin
        If (k Mod 8)<>1 Then s:=s+' ';
        s:=s+LongInt2HexN(2,ab[k]);
        If ((k Mod 8)=0) Or (k=nab) Then
          Begin
          Writeln(fl,s);
          Inc(pe);
          s:=ListGetAddressAlone;
          Dec(pe);
          End;
        End;
      Inc(pe);
      End;
    End;
  End;
End;

Procedure ProcessDw;
Var w:Word;
  s:String;
  k,naw:Byte;
  aw:Array[1..255] Of Word;
Begin
naw:=0;
For k:=1 To cl.nsp Do
  Begin
  If (Pos('''',cl.asp[k])>0) Or (Pos('"',cl.asp[k])>0) Then
    Error(49,'','') Else
    Begin
    GetConstantWord(cl.asp[k],w);
    Inc(naw);
    aw[naw]:=w;
    End;
  End;
If naw>0 Then
  Begin
  s:=ListGetAddressAlone;
  Case cSegm Of
    'C':For k:=1 To naw Do
      Begin
      If pass=2 Then HexWriteC(pc,aw[k]);
      Inc(pc);
      Inc(nconstants);
      If Not fListOff Then
        Begin
        If (k MOD 8)<>1 Then s:=s+' ';
        s:=s+LongInt2HexN(4,aw[k]);
        If ((k MOD 8)=0) Or (k=naw) Then
          Begin
          If Not fListOff Then Writeln(fl,s);
          s:=ListGetAddressAlone;
          End;
        End;
      End;
    'E':For k:=1 To naw Do
      Begin
      If pass=2 Then HexWriteE(pe,aw[k] MOD 256);
      Inc(pe);
      Inc(ne);
      If pass=2 Then HexWriteE(pe,aw[k] DIV 256);
      Inc(pe);
      Inc(ne);
      If Not fListOff Then
        Begin
        If (k MOD 4)<>1 Then s:=s+' ';
        s:=s+LongInt2HexN(2,aw[k] MOD 256)+' '+LongInt2HexN(2,aw[k] Div 256);
        If ((k MOD 4)=0) Or (k=naw) Then
          Begin
          If Not fListOff Then Writeln(fl,s);
          s:=ListGetAddressAlone;
          End;
        End;
      End;
    End;
  End Else
  Error(50,'','');
End;

Procedure ProcessMacro;
Begin
If (pass=1) And (Not WriteNewMacro(cl.asp[1])) Then
  Error(57,cl.asp[1],'');
fMacroDef:=True;
End;

Procedure ProcessEndMacro;
Begin
If (pass=1) And (Not WriteCloseMacro) Then
  Error(59,'','');
fMacroDef:=False;
End;

Procedure ProcessIf(fina:Boolean); { Call of routine in gavrif }
Var fValid:Boolean;
  i:LongInt;
Begin
fValid:=True;
If GetExpression(cl.asp[1],fValid,i) Then
  Begin
  If fValid Then IfNew(i<>1,fina) Else
    Error(61,'','');
  End Else
  Error(62,'','');
End;

Procedure ProcessIfDevice(fina:Boolean); { Call of routine in gavrif }
Var p:PSList;
Begin
If (cl.asp[1,1]='"') And (cl.asp[1,Length(cl.asp[1])]='"') Then
  Begin
  Delete(cl.asp[1],Length(cl.asp[1]),1);
  Delete(cl.asp[1],1,1);
  End;
If FindAvrType(cl.asp[1]) Then
  Begin
  p:=FindSymbolFirst('T');
  If p=NIL Then
    Error(7,cl.asp[1],'') Else
    IfNew(UpperCase(cl.asp[1])<>p^.sName,fina);
  End Else
  Error(92,cl.asp[1],'');
End;

Procedure ProcessEndIf; { call of routine in gavrif }
Begin
If Not IfEnd Then { .ENDIF }
  Error(64,'','');
End;

Procedure ProcessElse; { call of routine in gavrif }
Begin
If Not IfElse Then
  Error(65,'','');
End;

Procedure ProcessElif; { call of routine in gavrif }
Var fValid:Boolean;
  i:LongInt;
Begin
If IfElse Then
  Begin
  fValid:=True;
  If GetExpression(cl.asp[1],fValid,i) Then
    Begin
    If fValid Then
      fIfDef:=i<>1 Else
      Error(61,'','');
    End Else
    Error(62,'','');
  End Else
  Error(65,'','');
End;

Procedure ProcessIfDef(fina:Boolean); { call of routine in gavrif }
Var f:Boolean;
  i:LongInt;
Begin
f:=Not GetSymbolValue(sAllTypes,cl.asp[1],True,i);
IfNew(f,fina);
End;

Procedure ProcessIfNDef(fina:Boolean); { call of routine in gavrif }
Var f:Boolean;
  i:LongInt;
Begin
f:=GetSymbolValue(sAllTypes,cl.asp[1],True,i);
IfNew(f,fina);
End;

Procedure ProcessMessage;
Begin
Message(cl.sString);
Writeln;
End;

Procedure ProcessError;
Begin
Error(0,'','');
End;

Procedure ProcessExit;
Var fValid:Boolean;
  i:LongInt;
Begin
If cl.nsp=1 Then
  Begin
  fValid:=True;
  If GetExpression(cl.asp[1],fValid,i) Then
    Begin
    If fValid Then
      Begin
      fExit:=i=1;
      If fExit Then Error(86,'','');
      End Else
      Error(84,'','');
    End Else
    Error(85,'','');
  End Else
  fExitFile:=True;
End;

Procedure ProcessPseudo;
Var fe:Boolean;
Begin
If cl.nDirective>0 Then
  Begin
  sListLine2:='';
  If Not fListOff Then Writeln(fl,ListGetLine+cl.sLine);
  fe:=False;
  Case cSegm Of
    'C':Case cl.nDirective Of
          iDirByte:fe:=True;
          iDirEndMacro,iDirEndM:fe:=Not fMacroDef;
          End;
    'D':Case cl.nDirective Of
          iDirCseg,iDirDseg,iDirEseg,iDirOrg,iDirByte,iDirSet,iDirEqu,iDirError,
            iDirMessage,iDirInclude,iDirExit:;
	  Else
	  fe:=True;
          End;
    'E':Case cl.nDirective Of
          iDirCseg,iDirDseg,iDirEseg,iDirOrg,iDirDb,iDirDw,iDirError,
            iDirMessage,iDirSet,iDirEqu,iDirInclude,iDirExit:;
	  Else
	  fe:=True;
          End;
    End;
  If fe Then
    Error(66,cSegm,'') Else
    Case cl.nDirective Of
      iDirByte:ProcessByte; { .BYTE }
      iDirCseg:cSegm:='C'; { .CSEG }
      iDirDb:ProcessDb; { .DB }
      iDirDef:EquSetDef('R'); { .DEF }
      iDirDevice:ProcessDevice; { .DEVICE }
      iDirDseg:cSegm:='D'; { .DSEG }
      iDirDw:ProcessDw; { .DW }
      iDirElif:ProcessElif; { .ELIF}
      iDirElse:ProcessElse; { .ELSE }
      iDirEndIf:ProcessEndIf; { ENDIF }
      iDirEqu:EquSetDef('C'); { .EQU }
      iDirError:ProcessError; { .ERROR}
      iDirEseg:cSegm:='E'; { .ESEG }
      iDirExit:ProcessExit; { .EXIT }
      iDirIf:ProcessIf(False); { .IF }
      iDirIfDef:ProcessIfDef(False); { .IFDEF }
      iDirIfDevice:ProcessIfDevice(False); {.IFDEVICE }
      iDirIfNDef:ProcessIfNDef(False); { .IFNDEF }
      iDirInclude:IncludeFile; { .INCLUDE }
      iDirMessage:ProcessMessage; { .MESSAGE }
      iDirList:fListOff:=Not fListFileOpen; { .LIST }
      iDirListMac:fListMac:=True; { .LISTMAC }
      iDirMacro:ProcessMacro; { .MACRO }
      iDirEndMacro:ProcessEndMacro; { .ENDMACRO }
      iDirEndM:ProcessEndMacro; { .ENDM }
      iDirNoList:fListOff:=True; { .NOLIST }
      iDirOrg:ProcessOrg; { .ORG }
      iDirSet:EquSetDef('V'); { .SET }
      iDirSetGlobal:ProcessSetGlobal; { .SETGLOBAL }
      End;
  If (Not fListOff) And (sListLine2<>'') Then
    Writeln(fl,sListLine2);
  End Else
  Error(67,'','');
End;

{ ------------- }
{ Process lines }
{ ------------- }
Procedure ProcessMacroLine;
Var i:LongInt;
  s:String;
Begin
If (cl.sDirective='ENDMACRO') Or (cl.sDirective='ENDM') Then
  ProcessEndMacro Else
  If pass=1 Then With cl Do
  Begin
  If (sLabel<>'') And Not AddSymbol('M',sLabel,True,pcm^.ipcc) Then
      Error(55,sLabel,'');
  If sDirective='SET' Then
    Begin
    If Pos('=',asp[1])>0 Then
      Begin
      s:=Copy(asp[1],1,Pos('=',asp[1])-1);
      If snipoff(s) Then AddSymbol('V',s,False,0) Else
        Error(65,'','');
      End Else
      Error(65,'','');
    End;
  If sMacro<>'' Then { macro call within a macro }
    Begin
    If FindMacroLength(sMacro,i) Then
      Begin
      pcm^.ipcc:=pcm^.ipcc+i;
      pcm^.ipct:=pcm^.ipcc;
      End Else
      Error(84,sMacro,'');
    End;
  If Not WriteAddMacroLine(sLine) Then Error(68,'','');
  End;
End;

{$IFDEF debug}

Procedure DisplayMacros;
Var se:String;
Begin
If ResetMacroPointer Then
  Begin
  Writeln;
  Writeln(GetMsgM(11));
  Writeln(GetMsgM(12));
  While EnumMacroList(se) Do Writeln(se);
  End Else
  Writeln(GetMsgM(13));
End;

{$ENDIF}

Procedure ProcessInstructionLine; { normal instruction line }
Var i:LongInt;
Begin
If cl.sDirective<>'' Then
  ProcessPseudo Else With cl Do
  Begin
  If (sLabel<>'') Then { define/set label }
    Begin
    Case pass Of
      1:If Not AddSymbol('L',sLabel,True,ListGetAddressLongInt) Then
          Error(55,sLabel,'');
      2:If GetSymbolValue('LMG',sLabel,False,i) Then
        Begin
        If i<>ListGetAddressLongInt Then
          Dbg('Label value ($'+
            LongInt2HexN(6,ListGetAddressLongInt)+') <> pass1 ($'+
            LongInt2HexN(6,i)+')!');
        End Else
        Begin
        Dbg('Unexpected undefined label ('+sLabel+') in instruction line!');
        End;
      End;
    End;
  If sMnemo<>'' Then
    Begin { Process an instruction }
    AnalyseCommand;
    ListOutCommand;
    If(pass=2) Then
      Begin
      HexWriteC(pc,w1);
      If Inst.nW=2 Then HexWriteC(pc+1,w2);
      End;
    ncode:=ncode+Inst.nW;
    pc:=pc+Inst.nW;
    End Else
    Begin
    If sMacro<>'' Then
      Begin { Process a macro call }
{$IFDEF debug}     
      DisplayMacros;
{$ENDIF}
      If OpenReadMacro(sMacro,pc) Then
        Begin { Set macro parameter }
        If SetMacroParams(nsp,asp) Then
          ListOutLine Else
          Error(68,'','');
        End Else
        Error(70,'','');
      End Else
      ListOutLine;{ No instruction, no macro }
    End;
  End;
End;

Procedure ProcessLine;
Begin
If fMacroDef Or (fIf And fIfDef) Then { Inside macro def or if }
  Begin
  If Not fListOff Then
    Writeln(fl,ListGetLine,cl.sLine);
  If fMacroDef Then { line inside macro definition }
    Begin
    ProcessMacroLine;
    End Else With cl Do
    Begin { line inside if }
    Case nDirective Of
      iDirIf:ProcessIf(True);
      iDirIfDevice:ProcessIfDevice(True);
      iDirIfDef:ProcessIfDef(True);
      iDirIfNDef:ProcessIfNDef(True);
      iDirEndIf:ProcessEndIf;
      iDirElse:ProcessElse;
      iDirElif:ProcessElif;
      End;
    End;
  End Else
  ProcessInstructionLine;
End;

{ -------------- }
{ Process a file }
{ -------------- }
Function ProcessFile(sf:String):Boolean;
Var nl:LongInt;
  fs:Text;
  fh:Boolean;
  be:Byte;
Begin
If FileExists(sf) Then
  Begin
  fWithinDefInclude:=Pos('def.inc',sf)>0;
  spFile:=sf;
  nl:=0;
  nw:=0;
  Assign(fs,sf);
  Reset(fs);
  If Not fQuiet Then Writeln;
  While Not (Eof(fs) Or fExit Or fExitFile) Do { Process lines in file }
    Begin
    Inc(nl);
    If Not (fQuiet Or fAnsiOff) Then Writeln(Char(27),'[1A',GetMsgM(5),nl);
    npLine:=nl;
    Readln(fs,cl.sLine);
    If SplitL(be) Then
      Begin
      ProcessLine;
      fh:=fListOff;
      fListOff:=fListOff Or (Not fListMac);
      While fHasMacLines Do { Process macro lines, if any }
        Begin
        If ReadGetMacroLine(cl.sLine) Then
          Begin
          If cl.sLine<>'' Then
            Begin
            If SplitL(be) Then
              Begin
              cl.sLabel:='';
              ProcessLine;
              End Else
              Error(70+be,'','');
            End;
          End Else
          If fHasMacLines Then Error(80,'','');
        End;
      fListOff:=fh;
      End Else
      Error(70+be,'','');
    End;
  If Not fQuiet Then
    Begin
    If Not fAnsiOff Then Write(Char(27),'[1A');
    Writeln(nl,GetMsgM(16));
    Writeln;
    End;
  Close(fs);
  fWithinDefInclude:=False;
  fExitFile:=False;
  ProcessFile:=nerr=0;
  End Else
  Error(36,sf,'');
End;

{ ------------------------------------- }
{ Set all variables for starting a pass }
{ ------------------------------------- }
Procedure ClearForRestart;
Begin
pc:=0;
pe:=0;
pd:=$0060;
ncode:=0;
nconstants:=0;
ne:=0;
nd:=0;
cSegm:='C';
fExitFile:=False;
fMacroDef:=False;
fHasMacLines:=False;
ClearAllnUsed;
End;

Procedure OpenListFile; { Start list file }
Begin
fListOff:=fList;
If fListFileOpen Then Close(fl);
If Not fList Then
  Begin
  Assign(fl,sListFile);
  ReWrite(fl);
  Writeln(fl,lHeader);
  For k:=1 To Length(lHeader) Do Write(fl,'-');
  Writeln(fl);
  Writeln(fl,GetMsgM(17),sfn);
  Writeln(fl,GetMsgM(18),sCodeFile);
  Writeln(fl,GetMsgM(19),sEepFile);
  Writeln(fl,GetMsgM(20),GetDateAndTime);
  Writeln(fl,GetMsgM(21),pass);
  fListFileOpen:=True;
  End Else
  If FileExists(sListFile) Then DeleteFile(sListFile);
Assign(fx,sXErrorFile);
ReWrite(fx);
Writeln(fx,lHeader);
For k:=1 To Length(lHeader) Do Write(fx,'-');
Writeln(fx);
Writeln(fx,GetMsgM(17),sfn);
Writeln(fx,GetMsgM(18),sCodeFile);
Writeln(fx,GetMsgM(19),sEepFile);
Writeln(fx,GetMsgM(20),GetDateAndTime);
Writeln(fx,GetMsgM(21),pass);
End;

{ ----------- }
{ Do one pass }
{ ----------- }
Function DoPass:Boolean;
Var f:Boolean;
Begin
If Not fQuiet Then
  Begin
  Writeln('-------');
  Writeln(GetMsgM(21),pass);
  End;
OpenListFile;
If pass=2 Then
  HexOpenFiles(sCodeFile,sEepFile,ne);
ClearForRestart;
f:=ProcessFile(sfn);
With device Do
  Begin
  If pc>(nc Div 2) Then Error(81,IntToStr(pc),IntToStr(nc Div 2));
  If pe>ne Then Error(82,IntToStr(pe),IntToStr(ne));
  If (pd-$0060)>ns Then Warning(False,5,IntToStr(pd-$0060),IntToStr(ns));
  End;
If Not IfClear Then
  Error(99,'','');
If nErr>0 Then
  Writeln(fx,GetMsgM(35),nerr,GetMsgM(37));
Close(fx);
If nErr=0 Then DeleteFile(sXErrorFile);
DoPass:=f And (nErr=0);
End;

{ ------------ }
{ Main program }
{ ------------ }
Begin
For k:=1 To nHeader Do Writeln(asHeader[k]);
fList:=False;
fListOff:=False; { Set default values }
fListMac:=False;
fBeginner:=False;
{$IFDEF LINUX}
fAnsiOff:=False;
{$ELSE}
fAnsiOff:=True;
{$ENDIF}
fExit:=False;
With device Do
  Begin
  sn:='Undefined!';
  nc:=65535;
  ns:=65535;
  ne:=65535;
  ni:=131;
  End;
nw:=0;
If GetOptions Then { Get comman line options }
  Begin
  Writeln(GetMsgM(22),GetMsgM(17),sfn);
  nerr:=0;
  pass:=1;
  If DoPass Then { Do pass one }
    Begin
    Writeln('Pass 1 ok.');
    nw:=0;
    pass:=2;
    If DoPass Then { Do pass two }
      Begin
      If device.sn='Undefined!' Then
        Warning(False,6,'','');
      If Not CheckInstValid(hasDoc) Then
        Warning(False,10,'','');
      CheckSymbols;
      If Not fquiet Then { list results on screen }
        Begin
        Writeln;
        Writeln(ncode,GetMsgM(23),nconstants,GetMsgM(24),ncode+nconstants,' =',200.0*(ncode+nconstants)/device.nc:5:1,'%');
        If (device.ne>0) And (ne>0) Then
          Writeln(ne,' bytes EEPROM =',100.0*ne/device.ne:5:1,'%');
        If (device.ns>0) And (nd>0) Then
          Writeln(nd,' bytes SRAM =',100.0*nd/device.ns:5:1,'%.');
        Writeln;
        End;
      Case nw Of
        0:Writeln(GetMsgM(25));
        1:Writeln(GetMsgM(26));
        Else
        Writeln(nw,GetMsgM(27));
        End;
      Writeln(GetMsgM(28),GetMsgM(38));
      If Not fListOff Then { write results to list file }
        Begin
        If fSymbols Then
          Begin
          ListSymbols;
          ListMacros;
          End;
        Writeln(fl);
        Writeln(fl,GetMsgM(29),ncode:8,' words.');
        Writeln(fl,GetMsgM(30),nconstants:8,' words.');
        Writeln(fl,GetMsgM(31),ncode+nconstants:8,' words.');
        Writeln(fl,GetMsgM(32),ne:8,' bytes.');
        Writeln(fl,GetMsgM(33),nd:8,' bytes.');
        Writeln(fl,GetMsgM(28));
        Writeln(fl,GetMsgM(34),GetDateAndTime);
        End;
      HexCloseFiles;
      End;
    End;
  If nerr>0 Then
    Begin { List error messages }
    If nerr=1 Then
      Writeln(GetMsgM(35),GetMsgM(36)) Else
      Writeln(GetMsgM(35),nerr,GetMsgM(37));
    If Not fListOff Then
      Begin
      If fSymbols Then ListSymbols;
      If nerr=1 Then
        Writeln(fl,GetMsgM(35),GetMsgM(36)) Else
        Writeln(fl,GetMsgM(35),nerr,GetMsgM(37));
      End;
    HexDeleteFiles(sCodeFile,sEepFile);
    End;
  If fListFileOpen Then Close(fl); { Close list file }
  ClearSymbols; { Clear lists of symbols and macros }
  CloseAllMacros;
  End;
End.
