Unit gavriset;
{ Instruction set of AVRs, created by instset.pas }

Interface

Const
  { Definition of instruction set combinations }
  hasDoc:Word = $0001;
  hasAdiwSbiw:Word = $0002;
  hasIjmpIcall:Word = $0002;
  hasLdXY:Word = $0002;
  hasPushPop:Word = $0002;
  hasMulFmul:Word = $0004;
  hasJmpCall:Word = $0008;
  hasElpm:Word = $0010;
  hasElpmZ:Word = $0020;
  hasMovw:Word = $0040;
  hasLpmZ:Word = $0040;
  hasLpm:Word = $0080;
  hasSpm:Word = $0100;
  hasBreak:Word = $0200;
  hasEiJmpCall:Word = $0400;

 { Set the instruction set for the AVR type in s, }
 { returns true if type was found }
Function SetInstType(s:String):Boolean;

 { Check if the instruction type is fine in this AVR }
 { returns false if not valid }
Function CheckInstValid(w:Word):Boolean;

Implementation

Const
  nType = 52;
  aType:Array[1..nType] Of String = ( { AVR names }
        'AT90S1200','AT90S2313','AT90S2323','AT90S2333','AT90S2343','AT90S4414',
        'AT90S4433','AT90S4434','AT90S8515','AT90S8534','AT90S8535','ATTINY11',
        'ATTINY12','ATTINY13','ATTINY15','ATTINY22','ATTINY25','ATTINY26',
        'ATTINY28','ATTINY45','ATTINY85','ATTINY2313','ATMEGA103','ATMEGA104',
        'ATMEGA1280','ATMEGA1281','ATMEGA128','ATMEGA161','ATMEGA162','ATMEGA163',
        'ATMEGA165','ATMEGA168','ATMEGA169','ATMEGA16','ATMEGA2560','ATMEGA2561',
        'ATMEGA323','ATMEGA3250','ATMEGA325','ATMEGA32','ATMEGA48','ATMEGA603',
        'ATMEGA640','ATMEGA6450','ATMEGA645','ATMEGA64','ATMEGA83','ATMEGA88',
        'ATMEGA8515','ATMEGA8535','ATMEGA8','AT86RF401');

  aWt:Array[1..nType] Of Word = ( { AVR instruction combination }
      $0001,$0083,$0083,$0082,$0083,$0082,
      $0083,$0082,$0083,$0082,$0083,$0081,
      $0081,$03C3,$0081,$0083,$03C3,$0083,
      $0081,$03C3,$03C3,$03C3,$0093,$0092,
      $03FF,$03FF,$03FF,$01CF,$03CF,$01CF,
      $03CF,$03CF,$03CF,$03CF,$03FF,$03FF,
      $03CF,$03CF,$03CF,$03CF,$03C7,$0086,
      $03CF,$03CF,$03CF,$03CF,$0082,$03C7,
      $01CF,$03CF,$01CF,$00CB);

Var
  wType:Word; { Instruction capabilities of the AVR }

Function SetInstType(s:String):Boolean;
Var k:Byte;
Begin
SetInstType:=False;
k:=0;
Repeat Inc(k) Until (k>nType) Or (aType[k]=s);
If k<=nType Then
  Begin
  wType:=aWt[k];
  SetInstType:=True;
  End;
End;

Function CheckInstValid(w:Word):Boolean;
Begin
CheckInstValid:=(w And wType)=w;
End;

Begin
wType:=$FFFF;
End.
