{ characteristics of the AVR types for gavrasm }
Unit gavrtype;

Interface

Uses Crt;

Type TAvrType=Record { AVR type }
       sn:String[10]; { type name }
       nc:LongInt;{ program flash in bytes }
       ns:LongInt;{ SRAM in bytes }
       ne:LongInt;{ EEPROM in bytes }
       ni:Byte; { number of instructions }
       End;

Var device:TAvrType;

{ lists all AVR types and their characterica on screen }
Procedure ListAvrTypes;

{ find the AVR type in at.sn, set type and return true if found }
Function GetAvrType:Boolean;

{ find the AVR type in s and return true if found }
Function FindAvrType(s:String):Boolean;

Implementation

Uses SysUtils;

Const
  nAvrTypes=52;
  AvrType:Array[1..nAvrTypes] Of TAvrType = (
   (sn:'AT90S1200';nc:1024;ns:0;ne:64;ni:89),
   (sn:'AT90S2313';nc:2048;ns:128;ne:128;ni:118),
   (sn:'AT90S2323';nc:2048;ns:128;ne:128;ni:118),
   (sn:'AT90S2333';nc:2048;ns:128;ne:128;ni:118),
   (sn:'AT90S2343';nc:2048;ns:128;ne:128;ni:118),
   (sn:'AT90S4414';nc:2048;ns:256;ne:256;ni:118),
   (sn:'AT90S4433';nc:4096;ns:128;ne:256;ni:118),
   (sn:'AT90S4434';nc:4096;ns:256;ne:256;ni:118),
   (sn:'AT90S8515';nc:8192;ns:512;ne:512;ni:118),
   (sn:'AT90C8534';nc:8192;ns:256;ne:512;ni:118),
   (sn:'AT90S8535';nc:8192;ns:512;ne:512;ni:118),
   (sn:'ATtiny11';nc:1024;ns:0;ne:0;ni:90),
   (sn:'ATtiny12';nc:1024;ns:0;ne:64;ni:90),
   (sn:'ATtiny13';nc:1024;ns:64;ne:64;ni:120),
   (sn:'ATtiny15';nc:1024;ns:0;ne:64;ni:90),
   (sn:'ATtiny22';nc:2048;ns:128;ne:128;ni:118),
   (sn:'ATtiny25';nc:2048;ns:128;ne:128;ni:120),
   (sn:'ATtiny26';nc:2048;ns:128;ne:128;ni:118),
   (sn:'ATtiny28';nc:2048;ns:0;ne:0;ni:90),
   (sn:'ATtiny45';nc:4096;ns:256;ne:256;ni:120),
   (sn:'ATtiny85';nc:8192;ns:512;ne:512;ni:120),
   (sn:'ATtiny2313';nc:2048;ns:128;ne:128;ni:120),
   (sn:'ATmega103';nc:131072;ns:4096;ne:4096;ni:131),
   (sn:'ATmega104';nc:131072;ns:4096;ne:4096;ni:131),
   (sn:'ATmega1280';nc:131072;ns:8192;ne:4096;ni:133),
   (sn:'ATmega1281';nc:131072;ns:8192;ne:4096;ni:133),
   (sn:'ATmega128';nc:131072;ns:4096;ne:4096;ni:131),
   (sn:'ATmega161';nc:16384;ns:1024;ne:512;ni:130),
   (sn:'ATmega162';nc:16384;ns:1024;ne:512;ni:131),
   (sn:'ATmega163';nc:16384;ns:1024;ne:512;ni:130),
   (sn:'ATmega165';nc:16384;ns:1024;ne:512;ni:130),
   (sn:'ATmega168';nc:16384;ns:1024;ne:512;ni:131),
   (sn:'ATmega169';nc:16384;ns:1024;ne:512;ni:130),
   (sn:'ATmega16';nc:16384;ns:1024;ne:512;ni:131),
   (sn:'ATmega2560';nc:262144;ns:8192;ne:4096;ni:135),
   (sn:'ATmega2561';nc:262144;ns:8192;ne:4096;ni:135),
   (sn:'ATmega323';nc:32768;ns:2048;ne:1024;ni:130),
   (sn:'ATmega3250';nc:32768;ns:2048;ne:1024;ni:130),
   (sn:'ATmega325';nc:32768;ns:2048;ne:1024;ni:130),
   (sn:'ATmega32';nc:32768;ns:2048;ne:1024;ni:131),
   (sn:'ATmega48';nc:4096;ns:1024;ne:512;ni:130),
   (sn:'ATmega603';nc:65536;ns:4096;ne:2048;ni:131),
   (sn:'ATmega640';nc:65536;ns:8192;ne:4096;ni:130),
   (sn:'ATmega6450';nc:65536;ns:4096;ne:2048;ni:130),
   (sn:'ATmega645';nc:65536;ns:4096;ne:2048;ni:130),
   (sn:'ATmega64';nc:65536;ns:4096;ne:2048;ni:130),
   (sn:'ATmega83';nc:16384;ns:512;ne:512;ni:130),
   (sn:'ATmega8515';nc:8192;ns:512;ne:512;ni:130),
   (sn:'ATmega8535';nc:8192;ns:512;ne:512;ni:130),
   (sn:'ATmega88';nc:8192;ns:1024;ne:512;ni:130),
   (sn:'ATmega8';nc:8192;ns:1024;ne:512;ni:130),
   (sn:'AT86RF401';nc:1024;ns:128;ne:128;ni:118));

{ lists all AVR types and their characteristica on screen }
Procedure ListAvrTypes;
Var k,l:Byte;
  c:Char;
Begin
For k:=1 To nAvrTypes Do With AvrType[k] Do
  Begin
  If (k Mod 20)=1 Then
    Begin
    If k>1 Then
      Begin
      Write('  continue ... ');
      Repeat Until Keypressed;
      c:=ReadKey;
      End;
    Writeln;
    Writeln('List of supported AVR types and their properties');
    Writeln('Type        Flash   SRAM  EEPROM  Instructions');
    Writeln('            Bytes  Bytes   Bytes');
    End;
  Write(sn);
  For l:=Length(sn)+1 To 11 Do Write(' ');
  Writeln(nc:6,ns:7,ne:8,ni:14,'    ');
  End;
End;

{ find the AVR type in device.sn and return true if found }
Function GetAvrType:Boolean;
Var k:Byte;
Begin
k:=0;
device.sn:=UpperCase(device.sn);
Repeat Inc(k) Until (k>nAvrTypes) Or (device.sn=UpperCase(AvrType[k].sn));
GetAvrType:=k<=nAvrTypes;
If GetAvrType Then With AvrType[k] Do
  Begin
  device.sn:=sn;
  device.nc:=nc;
  device.ns:=ns;
  device.ne:=ne;
  device.ni:=ni;
  End;
End;

{ find the AVR type in s and return true if found }
Function FindAvrType(s:String):Boolean;
Var k:Byte;
Begin
s:=UpperCase(s);
k:=0;
Repeat Inc(k) Until (k>nAvrTypes) Or (s=UpperCase(AvrType[k].sn));
FindAvrType:=k<=nAvrTypes;
End;

Begin
End.
